@@ -1,4 +1,4 @@
-/* Copyright (c) 2009-2010 Sam Trenholme
+/* Copyright (c) 2009-2012 Sam Trenholme
  *
  * TERMS
  *
@@ -463,7 +463,7 @@ void dwc_process(dw_hash *cache, dw_str *query, uint8_t action) {
                 if(dwc_ttl_age(fetch,ttl) == -1) {
                         goto catch_dwc_process;
                 }
-                dwh_add(cache,query,fetch,ttl,1);
+                dwh_add(cache,query,fetch,-2,1);
                 dw_destroy(fetch);
                 fetch = 0;
         }
@@ -481,7 +481,7 @@ void dwc_process(dw_hash *cache, dw_str *query, uint8_t action) {
                 if(ttl == -1) {
                         goto catch_dwc_process;
                 }
-                dwh_add(cache,query,fetch,ttl,1);
+                dwh_add(cache,query,fetch,-2,1);
         }
 
 catch_dwc_process:
@@ -501,6 +501,7 @@ int dwc_check_ip(dwd_dict *blacklist_hash, dw_str *ip) {
                 return 0;
         }
 
+#ifndef NO_FILTER_RFC1918
         /* Filter private IPs; see http://crypto.stanford.edu/dns/ for
          * why this improves security */
         if(ip->len == 4 && key_n[DWM_N_filter_rfc1918] == 1) {
@@ -530,6 +531,7 @@ int dwc_check_ip(dwd_dict *blacklist_hash, dw_str *ip) {
                         return 1;
                 }
         }
+#endif /* NO_FILTER_RFC1918 */
 
         if(blacklist_hash == 0) {
                 return 0;
