@@ -1,4 +1,4 @@
-/* Copyright (c) 2002,2003 Sam Trenholme
+/* Copyright (c) 2002-2007 Sam Trenholme
  *
  * TERMS
  *
@@ -27,11 +27,16 @@
 #include <errno.h>
 #include <string.h>
 #include <time.h>
-#include <sys/types.h>
+#include <stdint.h>
+#ifndef MINGW32
 #include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#else /* MINGW32 */
+#include <winsock.h>
+#include <wininet.h>
+#endif /* MINGW32 */
 #include "../MaraDns.h"
 /* All of the labels */
 #include "askmara_labels_en.h"
@@ -40,12 +45,23 @@
 #include "../rng/rng-api-fst.h"
 /* END RNG USING CODE */
 
-int verbose_mode = 0; /* Whether to have short or verbose output; 
+#include "../libs/MaraHash.h"
+#include "../dns/functions_dns.h"
+#include "../parse/functions_parse.h"
+
+int verbose_output(js_string *uindata);
+int csv2_compatible_output(js_string *uindata, js_string *query);
+int out_answer(js_string *uindata,int *place);
+int csv2_answer(js_string *uindata, int *place, js_string *query);
+/* escape_stdout_csv2: From parse/Csv2_esc_txt.c and defined in
+   parse/Csv2_functions.h */
+extern int escape_stdout_csv2(js_string *js);
+
+int verbose_mode = 0; /* Whether to have short or verbose output;
                          1: verbose output; any other value: short output */
 int timeout = 21;
 
-/* So -Wall doesn't gag */
-extern int decomp_init();
+unsigned short dns_port = DNS_PORT;
 
 /* Octal escape any non-ASCII sequences in the askmara output; almost idential
    to show_esc_stdout */
@@ -61,12 +77,12 @@ int escape_stdout(js_string *js) {
 
     while(counter < js->unit_count) {
         this = *(js->string + counter);
-	if(this < 32 || this > 126) 
-	    printf("\\%03o",this);
-        else if(this == '\\') 
-	    printf("\\%c",this);
+        if(this < 32 || this > 126)
+            printf("\\%03o",this);
+        else if(this == '\\')
+            printf("\\%c",this);
         else
-	    printf("%c",this);
+            printf("%c",this);
         counter++;
         }
     return 1;
@@ -75,15 +91,14 @@ int escape_stdout(js_string *js) {
 /* Generate a psudo-random query-id based on the hostname we give it.
    This helps us test the server against a large number of query IDs */
 
-u_int16_t gen_id(char *hostname) {
+uint16_t gen_id(char *hostname) {
     /* BEGIN RNG USING CODE */
     /* May as well bring out the bug guns (e.g. our secure RNG) */
-    BYTE r_inBlock[17],r_outBlock[17],r_binKey[17];
-    BYTE r_keyMaterial[320];
+    MARA_BYTE r_inBlock[17],r_outBlock[17],r_binKey[17];
     keyInstance r_keyInst;
     cipherInstance r_cipherInst;
     unsigned char crypto_key[34];
-    int desc, len, counter;
+    int counter;
     char *mc1, *mc2;
     /* END RNG USING CODE */
     int ret = 4321;
@@ -94,34 +109,34 @@ u_int16_t gen_id(char *hostname) {
 
     /* Initialize the keys, including the "binKey" (is this used?) */
     memset(r_binKey,'0',16);
-    mc1 = "ñ";
+    mc1 = "";
     mc2 = mc1;
-    /* Set the crypto key to 'ññññññññññññññññ' (UTF-8 encoding) */
+    /* Set the crypto key to '' (UTF-8 encoding) */
     for(counter = 0; counter < 16; counter++) {
         *(crypto_key + counter) = *mc2;
-	mc2++;
-	if(*mc2 == 0) 
-	   mc2 = mc1;
+        mc2++;
+        if(*mc2 == 0)
+           mc2 = mc1;
         }
-    /*memset(crypto_key,'ñ',16);*/
+    /*memset(crypto_key,'',16);*/
     /* Make the first 16 characters of the hostname the key (padded by the
-       ñ character--do you see a n~ there?) */
-    strncpy(crypto_key,hostname,16);
-    /* Set the plaintext block to 'ññññññññññññññññ' (UTF-8 encoding) */
+        character--do you see a n~ there?) */
+    strncpy((char *)crypto_key,hostname,16);
+    /* Set the plaintext block to '' (UTF-8 encoding) */
     mc2 = mc1;
     for(counter = 0; counter < 16; counter++) {
         *(r_inBlock + counter) = *mc2;
-	mc2++;
-	if(*mc2 == 0) 
-	   mc2 = mc1;
+        mc2++;
+        if(*mc2 == 0)
+           mc2 = mc1;
         }
-    /*memset(r_inBlock,'ñ',16);*/
+    /*memset(r_inBlock,'',16);*/
     /* If the hostname is longer than 16 characters, make the plaintext
        the longer part of the hostname */
     if(strlen(hostname) > 16)
-    strncpy(r_inBlock,hostname+16,16);
+    strncpy((char *)r_inBlock,hostname+16,16);
 
-    if(makeKey(&r_keyInst, DIR_ENCRYPT, 128, crypto_key) != 1)
+    if(makeKey(&r_keyInst, DIR_ENCRYPT, 128, (char *)crypto_key) != 1)
         return JS_ERROR;
     if(cipherInit(&r_cipherInst, MODE_ECB, NULL) != 1)
         return JS_ERROR;
@@ -140,66 +155,93 @@ int harderror(char *msg) {
     exit(1);
     }
 
-main(int argc, char **argv) {
+int main(int argc, char **argv) {
     char *server_address = NULL;
-    struct sockaddr_in dns_udp, server;
+    struct sockaddr_in dns_udp;
+    struct sockaddr_in *pdns_udp = &dns_udp;
+#ifdef AUTHONLY
+    struct sockaddr_in6 dns_udp6;
+#endif
     int len_inet; /* Length */
-    int s; /* Socket */
-    int q; /* Used for binding, etc */
+    int s = -1; /* Socket */
     js_string *outdata; /* Outgoing data */
     js_string *indata, *uindata; /* Incoming data (uncompressed version) */
     q_header header; /* header data */
     q_question question;
-    int counter,place,count;
+    int place,count;
     fd_set rx_set; /* Using select() because if its timeout option */
     int maxd;      /* select() */
     struct timeval tv;  /* select() */
     int n; /* Select() return value */
     int nrd = 0; /* whether recursion is desired or not 1: No recursion;
-                   0: recursion desired */
+                    0: recursion desired */
     int id;
     int attempts = 10;
+    int desired_rr = -1; /* This is used to determine what RR type the
+                            user desires */
+    char *temp;
 
+#ifdef MINGW32
+    int err;
+    WSADATA wsaData;
+    WORD wVersionRequested = MAKEWORD(2,2);
+    err = WSAStartup(wVersionRequested, &wsaData);
+#endif /* MINGW32 */
     /* Determine what the query string is */
-    if(argc < 2 || argc > 6)
-        harderror(L_USAGE);
-
-    if(argc >= 3 && *argv[1] == '-' && *(argv[1] + 1) == 'n')
-        nrd = 1;
-
-    /* See if we are in verbose mode or not */
     verbose_mode = 0;
-    if(argc >= 3 + nrd && *argv[1 + nrd] == '-' && *(argv[1 + nrd] + 1) == 'v')
-        verbose_mode = 1;
-
-    /* See if we should allow a user defined timeout or not */
     timeout = 31;
-    if(argc >= 4 + nrd && *argv[1 + nrd] == '-' && *(argv[1 + nrd] + 1) == 
-       't') {
-         timeout = atoi(argv[2 + nrd]);
-	 verbose_mode = 2;
-	 }
+    server_address = ASKMARA_DEFAULT_SERVER;
+    argc--;
+    argv++;
+    while(argc > 0) {
+        if(argv[0] [0]== '-' && argv[0][1] == 'v')
+            verbose_mode = 1;
+        else if(argv[0][0] == '-' && argv[0][1] == 'n')
+            nrd = 1;
+        else if(argv[0][0] == '-' && argv[0][1] == 'p') {
+            if(argc < 2) {
+                harderror(L_USAGE);
+                }
+            argc--;
+            argv++;
+            dns_port = atoi(*argv);
+            if(dns_port < 1 || dns_port > 32000) {
+                    harderror("Wrong port number\n");
+                }
+            }
+        else if(argv[0][0] == '-' && argv[0][1] == 't') {
+            if(argc < 2) {
+                harderror(L_USAGE);
+                }
+            argc--;
+            argv++;
+            timeout = atoi(*argv);
+            if(timeout < 1)
+               harderror( L_INVALID_TIMEOUT);
+            verbose_mode = 2;
+            }
+        else if (argv[0][0] != '-')
+            break;
+        else
+            harderror(L_USAGE);
+        argc--;
+        argv++;
+        }
 
-    if(timeout < 1) 
-        harderror(L_INVALID_TIMEOUT);
+    /* we expect now the query string and possibly the server */
+    if(argc<1)
+        harderror(L_USAGE);
+    if(argc>1)
+        server_address = argv[1];
 
-    /* Determine what IP address to connect to */
-    if(argc == 3 + verbose_mode + nrd)
-        server_address = argv[2 + verbose_mode + nrd];
-    else
-        server_address = "127.0.0.3";
-
-    if((indata = js_create(512,1)) == 0)
+    if((indata = js_create(4512,1)) == 0)
        harderror(L_JS_CREATE_INDATA);
-    if((uindata = js_create(2048,1)) == 0)
+    if((uindata = js_create(12048,1)) == 0)
        harderror(L_JS_CREATE_UINDATA);
-    if((outdata = js_create(512,1)) == 0)
+    if((outdata = js_create(4512,1)) == 0)
        harderror(L_JS_CREATE_OUTDATA);
     if((question.qname = js_create(512,1)) == 0)
        harderror(L_JS_CREATE_QNAME);
-    /* Create a UDP client socket */
-    if((s = socket(AF_INET,SOCK_DGRAM,0)) == -1)
-        harderror(L_SOCKET);
 
     /* Create a socket address to use with sendto() */
     memset(&dns_udp,0,sizeof(dns_udp));
@@ -208,13 +250,33 @@ main(int argc, char **argv) {
        useful to be able to contact a MaraDNS server running on a high
        port number.  However, DNS_PORT should be 53 otherwise (defined
        in MaraDNS.h) */
-    dns_udp.sin_port = htons(DNS_PORT);
-    if((dns_udp.sin_addr.s_addr = inet_addr(server_address)) == INADDR_NONE)
-        harderror(L_MAL_IP);
-
+    dns_udp.sin_port = htons(dns_port);
     len_inet = sizeof(dns_udp);
 
-    id = gen_id(argv[1 + verbose_mode]);
+    if((dns_udp.sin_addr.s_addr = inet_addr(server_address)) == INADDR_NONE) {
+#ifdef AUTHONLY
+        memset(&dns_udp6,0,sizeof(dns_udp6));
+        dns_udp6.sin6_family = AF_INET6;
+        dns_udp6.sin6_port = htons(dns_port);
+        if( inet_pton(AF_INET6, server_address, &dns_udp6.sin6_addr) < 1) {
+#endif
+            harderror(L_MAL_IP);
+#ifdef AUTHONLY
+        } else {
+            len_inet = sizeof(dns_udp6);
+            pdns_udp = (struct sockaddr_in*)&dns_udp6;
+            if((s = socket(AF_INET6,SOCK_DGRAM,0)) == -1)
+                harderror(L_SOCKET);
+            }
+#endif
+    } else {
+         /* Create a UDP IPv4 client socket */
+        if((s = socket(AF_INET,SOCK_DGRAM,0)) == -1)
+            harderror(L_SOCKET);
+        }
+
+
+    id = gen_id(argv[0]);
     /* Format a DNS request */
     /* DNS header */
     header.id = id;
@@ -234,15 +296,43 @@ main(int argc, char **argv) {
     /* Create a DNS question , and put it in raw UDP format */
     /* DNS question -> looking up an A record... */
     question.qclass = 1; /* ...on the internet */
-    if(js_qstr2js(question.qname,argv[1 + verbose_mode + nrd]) == JS_ERROR)
+    /* We now allow people to specify any numeric record in this format:
+     * 40:t1.example.com., where 40 is the numeric rrtype */
+
+    temp = argv[0];
+    if(temp == 0) {
+            printf("Wrong number of arguments to askmara\n");
+            exit(43);
+    }
+    if(*temp >= '0' && *temp <= '9') {
+        int td = 0;
+        desired_rr = 0;
+        for(td = 0; td < 7; td++) {
+                if(*temp < '0' || *temp > '9') {
+                        break;
+                }
+                desired_rr *= 10;
+                desired_rr += *temp - '0';
+                temp++;
+        }
+        if(*temp != ':') {
+                printf("Invalid numeric rtype\n");
+                exit(44);
+        }
+        *temp = 'A';
+    }
+
+    if(js_qstr2js(question.qname,temp) == JS_ERROR)
         harderror(L_INVALID_Q); /* Invalid query */
 
     /* Make sure that the csv1-compatible output hashes all non-RR
        output */
     if(verbose_mode != 1)
         printf("# ");
-
-    printf("%s%s%s",L_QUERYING,server_address,L_NEWLINE); /* Querying the server with the IP... */
+    if(dns_port==DNS_PORT)
+        printf("%s%s%s",L_QUERYING,server_address,L_NEWLINE); /* Querying the server with the IP... */
+    else
+        printf("%s%s#%d%s",L_QUERYING,server_address,dns_port,L_NEWLINE); /* Querying the server with the IP... */
 
     /* Make 'Aexample.com.' raw UDP data */
 
@@ -250,6 +340,10 @@ main(int argc, char **argv) {
     if(place == JS_ERROR)
         harderror(L_INVALID_DQ); /* Invalid form of domain query */
     question.qtype = place;
+    /* 40:name.example.com. style queries */
+    if(desired_rr >= 0 && desired_rr <= 65535) {
+            question.qtype = desired_rr;
+    }
 
     /* Make a string containing the RAW UDP query */
     make_hdr(&header,outdata);
@@ -258,7 +352,7 @@ main(int argc, char **argv) {
     do {
         /* Send out a DNS request */
         if(sendto(s,outdata->string,outdata->unit_count,0,
-          (struct sockaddr *)&dns_udp,len_inet) < 0)
+          (struct sockaddr *)pdns_udp,len_inet) < 0)
            harderror(L_UDP_NOSEND); /* Unable to send UDP packet */
 
         /* Wait for a reply from the DNS server */
@@ -273,23 +367,24 @@ main(int argc, char **argv) {
         if(n == 0) /* Timeout */
             harderror("Timeout");
         if((count = recvfrom(s,indata->string,indata->max_count,0,
-                        (struct sockaddr *)&dns_udp,&len_inet)) < 0)
+                        (struct sockaddr *)pdns_udp,
+                        (socklen_t *)&len_inet)) < 0)
             harderror(L_DNS_R_ERROR); /* Problem getting DNS server response */
 
         indata->unit_count = count;
 
         /* Get the ID from the remote response */
         if(count > 2) {
-	   id = (*(indata->string) << 8) | *(indata->string + 1);
-	   }
+           id = (*(indata->string) << 8) | *(indata->string + 1);
+           }
         else {
-	   id = header.id + 1;
-	   }
+           id = header.id + 1;
+           }
 
         attempts--;
 
-	/* If the id doesn't match, get a reply again */
-	} while(id != header.id && attempts > 0); 
+        /* If the id doesn't match, get a reply again */
+        } while(id != header.id && attempts > 0);
 
     decomp_init(0);
     decompress_data(indata,uindata);
@@ -303,8 +398,9 @@ main(int argc, char **argv) {
     if(verbose_mode == 1)
         verbose_output(uindata);
     else
-        csv1_compatible_output(uindata,question.qname);
+        csv2_compatible_output(uindata,question.qname);
 
+    return 0;
     }
 
 /* Verbose form of askmara output
@@ -422,9 +518,9 @@ int out_answer(js_string *uindata,int *place) {
             printf("%s%s",L_NEWLINE,L_S_RNAME); /* SOA rname */
             escape_stdout(soa.rname);
             printf("%s%s%u%s",L_NEWLINE,L_S_SERIAL,soa.serial,L_NEWLINE);
-            printf("%s%d%s",L_S_REFRESH,soa.refresh,L_NEWLINE);
-            printf("%s%d%s",L_S_RETRY,soa.retry,L_NEWLINE);
-            printf("%s%d%s",L_S_EXPIRE,soa.expire,L_NEWLINE);
+            printf("%s%d%s",L_S_REFRESH,(int)soa.refresh,L_NEWLINE);
+            printf("%s%d%s",L_S_RETRY,(int)soa.retry,L_NEWLINE);
+            printf("%s%d%s",L_S_EXPIRE,(int)soa.expire,L_NEWLINE);
             printf("%s%u%s",L_S_MINIMUM,soa.minimum,L_NEWLINE);
             }
         else if(rr_hdr.type == RR_MX) {
@@ -456,7 +552,7 @@ int out_answer(js_string *uindata,int *place) {
                    L_DOT,*(uindata->string + *place + 2),
                    L_DOT,*(uindata->string + *place + 3),L_NEWLINE,L_NEWLINE);
             }
-        else if(rr_hdr.type = RR_TXT) {
+        else if(rr_hdr.type == RR_TXT) {
             printf("%s",L_TXT); /* Text String */
             read_txt(uindata,mx.exchange,*place);
             escape_stdout(mx.exchange);
@@ -467,15 +563,16 @@ int out_answer(js_string *uindata,int *place) {
 
         *place += rr_hdr.rdlength; /* To do: read the RD data itself for
                                       all MaraDNS supported RRs */
+        return 0;
         }
 
-/* csv1-compatible form of askmara output
+/* csv2-compatible form of askmara output
    input: Pointer to js_string object with uncompressed UDP data,
           Raw UDP for of the question
    output: None, shows data on standard output
 */
 
-int csv1_compatible_output(js_string *uindata, js_string *query) {
+int csv2_compatible_output(js_string *uindata, js_string *query) {
 
     int place,counter,count;
     q_question question;
@@ -489,6 +586,34 @@ int csv1_compatible_output(js_string *uindata, js_string *query) {
     if(read_hdr(uindata,&header) == JS_ERROR)
         harderror(L_INHEADER_CONV); /* Problem converting inheader */
 
+    if(header.tc == 1) {
+        printf("# Remote server said: TRUNCATED\n");
+    }
+
+    /* If not 0, show them the rcode from the remote server */
+    switch(header.rcode) {
+        case 1:
+                printf("# Remote server said: FORMAT ERROR\n");
+                break;
+        case 2:
+                printf("# Remote server said: SERVER FAILURE\n");
+                break;
+        case 3:
+                printf("# Remote server said: NAME ERROR\n");
+                break;
+        case 4:
+                printf("# Remote server said: NOT IMPLEMENTED\n");
+                break;
+        case 5:
+                printf("# Remote server said: REFUSED\n");
+                break;
+        case 0:
+                break;
+        default:
+                printf("# Non-RFC1035 RCODE %d sent from remote server\n",
+                       header.rcode);
+        }
+
     /* Move past the questions */
     place = 12;
     for(counter = 0;counter < header.qdcount;counter++) {
@@ -496,10 +621,10 @@ int csv1_compatible_output(js_string *uindata, js_string *query) {
         if(count < 0) /* Error handling */
             harderror(L_QERROR); /* Error reading question */
         printf("# Question: ");
-	if(hname_translate(question.qname,question.qtype) == JS_ERROR) 
-	    harderror(L_QERROR);
+        if(hname_translate(question.qname,question.qtype) == JS_ERROR)
+            harderror(L_QERROR);
         show_esc_stdout(question.qname);
-	printf("\n");
+        printf("\n");
         place += count;
         }
 
@@ -507,17 +632,17 @@ int csv1_compatible_output(js_string *uindata, js_string *query) {
 
     /* Get all of the an replies */
     for(counter = 0; counter < header.ancount; counter++)
-        csv1_answer(uindata,&place,query);
+        csv2_answer(uindata,&place,query);
 
     /* Get all of the ns replies */
     printf("# %s%s",L_NSREP,L_NEWLINE);
     for(counter = 0; counter < header.nscount; counter++)
-        csv1_answer(uindata,&place,query);
+        csv2_answer(uindata,&place,query);
 
     /* Get all of the ar replies */
     printf("# %s%s",L_ARREP,L_NEWLINE);
     for(counter = 0; counter < header.arcount; counter++)
-        csv1_answer(uindata,&place,query);
+        csv2_answer(uindata,&place,query);
 
     return 1;
     }
@@ -530,13 +655,15 @@ int csv1_compatible_output(js_string *uindata, js_string *query) {
    output: outputs to stdout the data in that string
 */
 
-int csv1_answer(js_string *uindata, int *place, js_string *query) {
+int csv2_answer(js_string *uindata, int *place, js_string *query) {
 
     q_rr rr_hdr;
     rr_soa soa;
     rr_mx mx;
     int count;
     js_string *lower1, *lower2;
+    js_string *t;
+    int rtype;
 
     /* Create the needed strings */
     if((rr_hdr.name = js_create(512,1)) == 0)
@@ -569,32 +696,32 @@ int csv1_answer(js_string *uindata, int *place, js_string *query) {
     if(js_issame(lower1,lower2) != 1)
         printf("#");
 
+    rtype = rr_hdr.type;
     hname_translate(rr_hdr.name,rr_hdr.type);
-    escape_stdout(rr_hdr.name);
-    /* We also show the type if this is an unsupported record type */
-    if(rr_hdr.type != RR_A &&
-       rr_hdr.type != RR_NS &&
-       rr_hdr.type != RR_CNAME &&
-       rr_hdr.type != RR_SOA &&
-       rr_hdr.type != RR_PTR &&
-       rr_hdr.type != RR_MX &&
-       rr_hdr.type != RR_TXT)
-        printf("|%d",rr_hdr.type);
-    printf("|%u",rr_hdr.ttl);
+    t = js_create(256,1);
+    if(js_copy(rr_hdr.name,t) == JS_ERROR) {
+            harderror("Couldn't copy hostname");
+    }
+    t->string++;
+    t->unit_count--;
+    escape_stdout(t);
+    t->string--;
+    t->unit_count++;
+    js_destroy(t);
     if(rr_hdr.type == RR_SOA) {
         if(read_soa(uindata,&soa,*place) == JS_ERROR)
             harderror(L_READSOA);
-        hname_translate(soa.mname,RR_SOA);
-        hname_translate(soa.rname,RR_MX);
-        printf("|",L_S_MNAME); /* SOA mname */
+        hname_translate(soa.mname,RR_MAGIC_SPACE);
+        hname_translate(soa.rname,RR_MAGIC_EMAIL);
+        printf(" +%u",rr_hdr.ttl);
+        printf(" soa"); /* SOA mname */
         escape_stdout(soa.mname);
-        printf("|",L_NEWLINE,L_S_RNAME); /* SOA rname */
         escape_stdout(soa.rname);
-        printf("|%u",soa.serial,L_NEWLINE);
-        printf("|%d",soa.refresh,L_NEWLINE);
-        printf("|%d",soa.retry,L_NEWLINE);
-        printf("|%d",soa.expire,L_NEWLINE);
-        printf("|%u\n",soa.minimum,L_NEWLINE);
+        printf(" %u",soa.serial);
+        printf(" %d",(int)soa.refresh);
+        printf(" %d",(int)soa.retry);
+        printf(" %d",(int)soa.expire);
+        printf(" %u\n",soa.minimum);
         }
     else if(rr_hdr.type == RR_MX) {
         if(uindata->unit_count < *place + 2)
@@ -602,15 +729,28 @@ int csv1_answer(js_string *uindata, int *place, js_string *query) {
         mx.preference = ((*(uindata->string + *place) & 0xff) << 8) |
                          (*(uindata->string + *place + 1) & 0xff);
         read_ns(uindata,mx.exchange,*place + 2);
-        printf("|%d",mx.preference,L_NEWLINE); /* MX pref. */
-        hname_translate(mx.exchange,-2); /* Makes the char a pipe */
+        printf(" +%u mx %d",rr_hdr.ttl,mx.preference); /* MX pref. */
+        hname_translate(mx.exchange,RR_MAGIC_SPACE); /* Makes the char a
+                                                      * space */
         escape_stdout(mx.exchange);
         printf("\n");
         }
     else if(rr_hdr.type == RR_NS || rr_hdr.type == RR_CNAME ||
             rr_hdr.type == RR_PTR) {
+        printf(" +%u",rr_hdr.ttl);
+        switch(rr_hdr.type) {
+                case RR_NS:
+                        printf(" ns");
+                        break;
+                case RR_CNAME:
+                        printf(" cname");
+                        break;
+                case RR_PTR:
+                        printf(" ptr");
+                        break;
+                }
         read_ns(uindata,mx.exchange,*place);
-        hname_translate(mx.exchange,-2); /* First char is a pipe */
+        hname_translate(mx.exchange,RR_MAGIC_SPACE); /* First char is space */
         escape_stdout(mx.exchange);
         printf("\n");
         }
@@ -618,22 +758,66 @@ int csv1_answer(js_string *uindata, int *place, js_string *query) {
         if(uindata->unit_count < *place + 4)
             return JS_ERROR;
         /* Display the IP of the data */
-        printf("|%d%s%d%s%d%s%d\n",*(uindata->string + *place),
+        printf(" +%u a %d%s%d%s%d%s%d\n", rr_hdr.ttl,
+                        *(uindata->string + *place),
                L_DOT,*(uindata->string + *place + 1),
                L_DOT,*(uindata->string + *place + 2),
                L_DOT,*(uindata->string + *place + 3));
         }
-    else if(rr_hdr.type = RR_TXT) {
-        printf("|"); 
-        read_txt(uindata,mx.exchange,*place);
-        escape_stdout(mx.exchange);
-        printf("\n");
+    else if(rr_hdr.type == RR_AAAA) {
+        unsigned short *p;
+        if(uindata->unit_count < *place + 16)
+            return JS_ERROR;
+        /* Display the IP of the data */
+        p = (unsigned short*)(uindata->string + *place);
+        printf(" +%u aaaa %x:%x:%x:%x:%x:%x:%x:%x\n", rr_hdr.ttl,
+                   htons(*(p + 0 )),
+                   htons(*(p + 1)),
+                   htons(*(p + 2)),
+                   htons(*(p + 3)),
+                   htons(*(p + 4)),
+                   htons(*(p + 5)),
+                   htons(*(p + 6)),
+                   htons(*(p + 7)));
         }
+    else if(rr_hdr.type == RR_TXT) {
+        int len,p;
+        len = rr_hdr.rdlength;
+        p = *place;
+        printf(" +%u txt ",rr_hdr.ttl);
+        /* XXX: This doesn't correctly handle multi-chunked TXT data */
+        while(len > 0) {
+            int q;
+            q = read_txt(uindata,mx.exchange,p);
+            if(q < 0) {
+                printf("Problem reading TXT record\n");
+                exit(1);
+                }
+            q++;
+            escape_stdout_csv2(mx.exchange);
+            len -= q;
+            p += q;
+            if(len > 0) {
+                printf(";");
+                }
+            else {
+                printf("\n");
+                }
+            }
+         }
     else {
-        printf("|\n");
-	}
+        printf(" +%u raw %d ",rr_hdr.ttl,rr_hdr.type);
+        if(js_substr(uindata,mx.exchange,*place,rr_hdr.rdlength) == JS_ERROR) {
+                printf("Problem copying string over\n");
+                exit(55);
+        }
+        escape_stdout_csv2(mx.exchange);
+        printf(" \n");
+        }
 
     *place += rr_hdr.rdlength; /* To do: read the RD data itself for
                                   all MaraDNS supported RRs */
+    return 0;
     }
 
+
